! WHIZARD 1.00beta Dec 01 2000
! 
! (C) 1999, 2000 by Wolfgang Kilian <kilian@particle.uni-karlsruhe.de>
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module decay_forests

  use kinds, only: default !NODEP!
  use limits, only: MAX_EXTERNAL, MAX_CHANNELS, BUFFER_SIZE
  use mpi90, only: mpi90_broadcast !NODEP!
  use particles
  use decay_trees, decay_tree_TC => TC
  use mappings
  use file_utils !NODEP!
  use parser
  use cascades, only: make_phase_space_file
  use parameters, only: parameter_set !NODEP!

  implicit none
  private

  integer, parameter, public :: TC = decay_tree_TC

  type, public :: decay_forest
     integer :: n_externals, n_internals, n_in, n_out
     integer :: n_channels_max, n_channels, n_groves
     integer :: n_dim, n_dim0, n_dim1, n_dim2
     integer(kind=TC) :: n_branches_tot, n_branches_decay
     real(kind=default), dimension(:), pointer :: mass_sum
     type(decay_tree), dimension(:), pointer :: tree
     type(mapping), dimension(:,:), pointer :: map
     type(decay_grove), dimension(:), pointer :: grove
  end type decay_forest
  type, public :: decay_grove
     integer :: first, last
  end type decay_grove

  public :: create, destroy
  public :: write
  public :: read
  public :: forest_count
  public :: add
  public :: map
  public :: mpi90_broadcast
  public :: forest_set_strfun
  public :: forest_prepare

  interface create
     module procedure forest_create
  end interface
  interface destroy
     module procedure forest_destroy
  end interface
  interface write
     module procedure forest_write_unit
  end interface
  interface read
     module procedure forest_read_unit, forest_read_name
  end interface
  interface forest_count
     module procedure forest_count_name, forest_count_unit
  end interface
  interface add
     module procedure forest_add_tree
  end interface
  interface map
     module procedure forest_set_mapping
  end interface
  interface mpi90_broadcast
     module procedure forest_broadcast
  end interface

contains

  subroutine forest_create(f, n_in, n_out, n_channels_max)
    type(decay_forest), intent(out) :: f
    integer, intent(in) :: n_in, n_out, n_channels_max
    !allocate(f)
    f%n_in = n_in
    f%n_out = n_out
    f%n_externals = n_in + n_out
    f%n_channels_max  = n_channels_max
    f%n_channels  = 0
    f%n_groves = 0
    f%n_internals = f%n_externals-3
    f%n_branches_tot = 2**f%n_externals-1
    f%n_branches_decay = 2**(n_out)-1
    f%n_dim0  = 0
    f%n_dim1  = n_out - 2
    f%n_dim2  = 2*n_out - 2
    f%n_dim = f%n_dim0 + f%n_dim1 + f%n_dim2
    allocate(f%mass_sum(f%n_branches_tot))
    allocate(f%tree(f%n_channels_max))
    allocate(f%map(f%n_branches_tot, f%n_channels_max))
    f%map = mapping_null
    allocate(f%grove(f%n_channels_max))
  end subroutine forest_create

  subroutine forest_destroy(f)
    type(decay_forest), intent(inout) :: f
    integer :: ch
    deallocate(f%grove)
    deallocate(f%map)
    do ch=1, f%n_channels
       call destroy(f%tree(ch))
    end do
    deallocate(f%tree)
    deallocate(f%mass_sum)
    !deallocate(f)
  end subroutine forest_destroy

  subroutine forest_write_unit(u,f)
    integer, intent(in) :: u
    type(decay_forest), intent(in) :: f
    integer(kind=TC) :: k
    integer :: ch, igrove
    write(u,'(1x,A,1x,I3)') 'Overall dimension:    ', f%n_dim
    write(u,'(1x,A,1x,I3)') 'Phase space dimension:', f%n_dim1+f%n_dim2
    write(u,'(1x,A)') 'Phase space channels:'
    igrove = 1
    do ch=1, f%n_channels
       if (igrove <= f%n_groves) then
          if (ch == f%grove(igrove)%first) &
               & write(u, '(1x,A,I3)') 'Decay grove #', igrove
       end if
       write(u,'(1x,A,I3)') 'Decay tree #', ch
       call write(u,f%tree(ch))
       do k=1, f%n_branches_tot
          call write(u, f%map(k,ch), k)
       end do
       if (igrove <= f%n_groves) then
          if (ch == f%grove(igrove)%last) then
             write(u, '(1x,A)') 'End of grove'
             igrove = igrove+1
          end if
       end if
       write(u,*)
    end do
  end subroutine forest_write_unit

  subroutine forest_read_name(name, f, process_id)
    character(len=*), intent(in) :: name
    type(decay_forest), intent(inout) :: f
    character(len=*), intent(in) :: process_id
    integer :: unit
    logical :: exist
    inquire(file=name, exist=exist)
    if (exist) then
       unit = free_unit()
       open(unit=unit, file=name, status='old', action='read')
       call forest_read_unit(unit, f, process_id)
       call lex_clear
       close(unit)
    else
       print *, name
       stop 'Phase space configuration file not found'
    end if
  end subroutine forest_read_name

  subroutine forest_read_unit(unit, f, process_id)
    integer, intent(in) :: unit
    type(decay_forest), intent(inout) :: f
    character(len=*), intent(in) :: process_id
    integer :: iostat
    integer, dimension(2*MAX_EXTERNAL) :: a
    integer :: i, ch, code, mode
    integer(kind=TC) :: k
    logical :: ok, within_grove
    character(len=BUFFER_SIZE) :: string
    character(len=6) :: name
    type(mapping), dimension(f%n_branches_tot) :: default_map
    default_map = mapping_null
    FIND_PROCESS: do
       call find_string(unit, 'process', ok=ok)
       if (.not.ok) exit FIND_PROCESS
       call find_string(unit, process_id, &
            & reject=(/ 'process  ', 'grove    ', 'tree     ', &
            &           'cut      ', 'helicity ', 'flavor   ', &
            &           'histogram', 'and      ' /), &
            & ok=ok)
       if (ok) exit FIND_PROCESS
    end do FIND_PROCESS
    if (.not.ok) &
         & stop 'No appropriate process entry found in phase space configuration file'
    SCAN_DEFAULT_MAPPINGS: do
       call find_string(unit, 'map', reject=(/'process', 'grove  ', 'tree   '/), &
            &           ok=ok, &
            &           iostat=iostat)
       if (.not.ok) exit SCAN_DEFAULT_MAPPINGS
       k = get_integer(unit)
       string = get_string(unit)
       select case(string(1:1))
       case('s','S');  mode = S_CHANNEL
       case('t','T');  mode = T_CHANNEL
       case('u','U');  mode = U_CHANNEL
       case('o','O');  mode = ON_SHELL
       case('n','-');  mode = NO_MAPPING
       case default;   call parse_error(string)
       end select
       name = get_string(unit)
       code = particle_code(name)
       if (mode/=NO_MAPPING .and. code==UNDEFINED) call parse_error(string)
       call set(default_map(k), mode=mode, code=code)
    end do SCAN_DEFAULT_MAPPINGS
    within_grove = .false.
    SCAN_TREES: do ch=1, MAX_CHANNELS
       call find_string(unit, 'grove', reject=(/'process', 'tree   '/), &
            &           ok=ok, iostat=iostat)
       if (ok .or. ch > f%n_channels_max) then
          call find_string(unit, 'end', reject=(/'process', 'tree   '/), &
               &           ok=ok, iostat=iostat)
          if (ok .or. ch > f%n_channels_max) then
             within_grove = .false.
          else
             f%n_groves = f%n_groves + 1
             f%grove(f%n_groves)%first = ch
             within_grove = .true.
          end if
       end if
       if (within_grove) f%grove(f%n_groves)%last = ch
       call find_string(unit, 'tree', reject=(/'process'/), ok=ok, &
            &           iostat=iostat)
       if (.not.ok) exit SCAN_TREES
       SET_TREE_CODES: do i=1, size(a)
          a(i) = get_integer(unit, ok=ok)
          if (.not.ok) exit SET_TREE_CODES
       end do SET_TREE_CODES
       call forest_add_tree(f, a(:i-1), default_map)
       SCAN_MAPPINGS: do
          call find_string(unit, 'map', here=.true., ok=ok, &
               &          iostat=iostat)
          if (iostat < 0) exit SCAN_TREES
          if (.not.ok) exit SCAN_MAPPINGS
          k = get_integer(unit)
          string = get_string(unit)
          select case(string(1:1))
          case('s','S');  mode = S_CHANNEL
          case('t','T');  mode = T_CHANNEL
          case('u','U');  mode = U_CHANNEL
          case('o','O');  mode = ON_SHELL
          case('n','-');  mode = NO_MAPPING
          case default;   call parse_error(string)
          end select
          name = get_string(unit)
          code = particle_code(name)
          if (mode/=NO_MAPPING .and. code==UNDEFINED) call parse_error(string)
          call forest_set_mapping(f, k, mode, code, ch)
       end do SCAN_MAPPINGS
    end do SCAN_TREES
  end subroutine forest_read_unit

  subroutine forest_count_name(name, process_id, code, par, n_channels, &
       & generate, read_file, overwrite, &
       & keep_single, keep_double, keep_branchings, keep_all, keep_massive, &
       & off_shell, exchange, extra, s_channel)
    character(len=*), intent(in) :: name
    character(len=*), intent(in) :: process_id
    integer, dimension(:,:), intent(in) :: code
    type(parameter_set), intent(in) :: par
    integer, intent(out) :: n_channels
    logical, intent(in) :: generate, read_file, overwrite, s_channel
    logical, intent(in) :: keep_single, keep_double, keep_branchings
    logical, intent(in) :: keep_all, keep_massive
    integer, intent(in) :: off_shell, exchange, extra
    integer :: unit
    logical :: exist
    n_channels = 0
    inquire(file=name, exist=exist)
    if (read_file) then
       if (exist) then
          unit = free_unit()
          print *, '! Reading phase space information from file ', name, ' ...'
          open(unit=unit, file=name, status='old', action='read')
          call forest_count_unit(unit, process_id, n_channels)
          call lex_clear
          close(unit)
       else
          print *, '! Phase space configuration file ', name, ' not found'
       end if
    end if
    if (n_channels==0 .and. exist .and. generate .and. .not.overwrite) then
       print *, "I can try to generate a phase space configuration, but I don't want to"
       print *, "overwrite ", name, ".  You may either remove this file or set"
       print *, "overwrite_phase_space=T to overwrite it."
       stop
    end if
    if (n_channels==0) then
       if (.not.(generate)) &
            & stop "I don't know how to integrate this process, sorry."
       call make_phase_space_file (name, process_id, code, par, n_channels, &
            & keep_single, keep_double, keep_branchings, &
            & keep_all, keep_massive, &
            & off_shell=off_shell, peripheral=exchange, extra=extra, &
            & s_channel=s_channel, &
            & screen=.true.)
    end if
  end subroutine forest_count_name

  subroutine forest_count_unit(unit, process_id, n_channels)
    integer, intent(in) :: unit
    character(len=*), intent(in) :: process_id
    integer, intent(out) :: n_channels
    integer :: iostat
    integer :: ch
    logical :: ok
    character(len=BUFFER_SIZE) :: string
    n_channels = 0
    FIND_PROCESS: do
       call find_string(unit, 'process', ok=ok)
       if (.not.ok) exit FIND_PROCESS
       call find_string(unit, process_id, &
            & reject=(/ 'process  ', 'grove    ', 'tree     ', &
            &           'cut      ', 'helicity ', 'flavor   ', &
            &           'histogram', 'and      ' /), &
            & ok=ok)
       if (ok) exit FIND_PROCESS
    end do FIND_PROCESS
    if (.not.ok) then
       print *, "! No entry found for process ", process_id, "."
       return
    end if
    SCAN_TREES: do ch=1, MAX_CHANNELS
       call find_string(unit, 'tree', reject=(/'process'/), ok=ok, &
            &           iostat=iostat)
       if (.not.ok) exit SCAN_TREES
       n_channels = n_channels+1
       SCAN_MAPPINGS: do
          call find_string(unit, 'map', here=.true., ok=ok, &
               &          iostat=iostat)
          if (iostat < 0) exit SCAN_TREES
          if (.not.ok) exit SCAN_MAPPINGS
          string = get_string(unit)
          string = get_string(unit)
          string = get_string(unit)
       end do SCAN_MAPPINGS
    end do SCAN_TREES
    if (n_channels>0) then
       print *, "! ", n_channels, " phase space channels found for process ", &
            & process_id
    else
       print *, "! No phase space channels found for process ", process_id
       stop "Don't know how to integrate this process, sorry."
    end if
  end subroutine forest_count_unit

  subroutine forest_add_tree(f, a, default_map)
    type(decay_forest), intent(inout) :: f
    integer(kind=TC), dimension(:), intent(in) :: a
    type(mapping), dimension(:), intent(in), optional :: default_map
    integer :: i,ch
    ch = f%n_channels+1
    if (ch>f%n_channels_max) stop "Maximum number of decay channels exceeded"
    call create(f%tree(ch), f%n_in, f%n_out)
    call tree_from_array(f%tree(ch), a)
    call flip_t_to_s_channel(f%tree(ch))
    if (present(default_map)) then
       do i=1, size(a)
          f%map(flipped(f%tree(ch),a(i)), ch) = default_map(a(i))
       end do
    end if
    f%n_channels = ch
  end subroutine forest_add_tree

  subroutine forest_set_mapping(f, k, mode, code, channel)
    type(decay_forest), intent(inout) :: f
    integer(kind=TC), intent(in) :: k
    integer, intent(in) :: mode, code
    integer, intent(in), optional :: channel
    integer :: ch
    ch = f%n_channels;  if (present(channel)) ch = channel
    if (ch>f%n_channels) stop 'Attempt to define mapping for absent channel'
    call set(f%map(flipped(f%tree(ch),k),ch), mode=mode, code=code)
  end subroutine forest_set_mapping

  subroutine forest_broadcast(f, root, domain, error)
    type(decay_forest), intent(inout) :: f
    integer, intent(in) :: root
    integer, intent(in), optional :: domain
    integer, intent(out), optional :: error
    integer, dimension(:), allocatable :: buffer
    allocate(buffer(size(transfer(f, buffer))))
    buffer = transfer(f, buffer)
    call mpi90_broadcast(buffer, root, domain, error)
    f = transfer(buffer, f)
    deallocate(buffer)
  end subroutine forest_broadcast

  subroutine forest_set_strfun(f, n_strfun)
    type(decay_forest), intent(inout) :: f
    integer, intent(in) :: n_strfun
    f%n_dim0 = n_strfun
    f%n_dim = f%n_dim0 + f%n_dim1 + f%n_dim2
  end subroutine forest_set_strfun

  subroutine forest_prepare(f, code, par)
    type(decay_forest), intent(inout) :: f
    integer, dimension(:), intent(in) :: code
    type(parameter_set), intent(in) :: par
    integer(kind=TC) :: k
    integer :: i, ch
    f%mass_sum = 0
    do k=1, f%n_branches_decay
       do i=0, ubound(code,1)-1
          if (btest(k,i)) then
             if (ibclr(k,i)==0) then
                f%mass_sum(k) = particle_mass(code(i+1), par)
             else
                f%mass_sum(k) = f%mass_sum(ibclr(k,i))+f%mass_sum(ibset(0,i))
             end if
             exit
          end if
       end do
       do ch=1, f%n_channels
          call set(f%map(k,ch), par, k)
       end do
    end do
  end subroutine forest_prepare


end module decay_forests
