! WHIZARD 1.00beta Dec 01 2000
! 
! (C) 1999, 2000 by Wolfgang Kilian <kilian@particle.uni-karlsruhe.de>
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module decay_trees

  use limits, only: MAX_EXTERNAL

  implicit none
  private

  integer, parameter, public :: TC = selected_int_kind(MAX_EXTERNAL)

  type, public :: decay_tree
     integer :: n_branches, n_externals, n_in
     integer(kind=TC) :: mask, mask_in, mask_out
     type(branch), dimension(:), pointer :: branch
  end type decay_tree

  type, public :: branch
     logical :: set
     integer(kind=TC) :: mother, sibling, friend, origin
     integer(kind=TC), dimension(2) :: daughter
  end type branch

  public :: create, destroy
  public :: write
  public :: decay_level
  public :: tree_from_array
  public :: tree_to_array
  public :: flip_t_to_s_channel
  public :: flipped

  interface create
     module procedure tree_create
  end interface
  interface destroy
     module procedure tree_destroy
  end interface
  interface write
     module procedure tree_write_unit
  end interface

contains

  subroutine tree_create(t, n_in, n_out)
    type(decay_tree), intent(out) :: t
    integer, intent(in) :: n_in, n_out
    integer(kind=TC) :: i,n
    integer :: j
    !allocate(t)
    n = 2**(n_in+n_out)-1
    t%n_externals = n_in+n_out
    t%mask = 0
    do i=0, n_in+n_out-1
       t%mask = ibset(t%mask, i)
    end do
    t%n_in = n_in
    t%mask_in = 0
    do i=n_out, n_in+n_out-1
       t%mask_in = ibset(t%mask_in, i)
    end do
    t%mask_out = ieor(t%mask, t%mask_in)
    allocate(t%branch(n))
    do i=1, n
       call branch_create(t%branch(i))
    end do
    t%n_branches  = 0
  contains
    subroutine branch_create(b)
      type(branch), intent(out) :: b
      b%set = .false.
      do j=1,2
         b%daughter(j) = 0
      end do
      b%mother  = 0
      b%sibling = 0
      b%friend  = 0
      b%origin  = 0
    end subroutine branch_create
  end subroutine tree_create

  subroutine tree_destroy(t)
    type(decay_tree), intent(inout) :: t
    deallocate(t%branch)
    !deallocate(t)
  end subroutine tree_destroy

  subroutine tree_write_unit(u,t)
    integer, intent(in) :: u
    type(decay_tree), intent(in) :: t
    integer(kind=TC) :: k
    write(u,'(1X,A,I2,5X,A,I3)') &
         &'External:', t%n_externals, 'Mask:', t%mask
    write(u,'(1X,A,I2,5X,A,I3)') &
         &'Incoming:', t%n_in, 'Mask:', t%mask_in
    write(u,'(1X,A,I2,5X,A,I3)') &
         &'Branches:', t%n_branches
    do k = size(t%branch), 1, -1
       if (t%branch(k)%set) call branch_write_unit(u,t%branch(k))
    end do
  contains
    subroutine branch_write_unit(u,b)
      integer, intent(in) :: u
      type(branch), intent(in) :: b
      character(len=6) :: tmp
      if (b%origin /= 0) then
         write(tmp, '(A,I4,A)') '(', b%origin, ')'
      else
         tmp = ' '
      end if
      if (b%daughter(1) /= 0) then
         if (b%friend /= 0) then
            write(u,'(1X,A,I4,1x,A,2X,A,I4,I4,6X,A,I4)') &
                 &   '*', k, tmp, &
                 &   'Daughters: ', b%daughter(1), b%daughter(2), &
                 &   'Friend:    ', b%friend
         else
            write(u,'(1X,A,I4,1x,A,2X,A,I4,I4)') &
                 &   '*', k, tmp, &
                 &   'Daughters: ', b%daughter(1), b%daughter(2)
         end if
      else
         write(u,'(2X,I4,3X,A,I4,I4)') k
      end if
    end subroutine branch_write_unit
  end subroutine tree_write_unit

  function decay_level(k) result(l)
    integer(kind=TC) :: k
    integer :: l
    integer :: i
    l = 0
    do i=0, bit_size(k)-1
       if (btest(k,i)) l = l+1
    end do
  end function decay_level

  subroutine tree_from_array(t,a)
    type(decay_tree), intent(inout) :: t
    integer(kind=TC), dimension(:), intent(in) :: a
    integer :: i
    integer(kind=TC) :: k
    do i=1, size(a)
       k = a(i)
       if (iand(k, t%mask_in) == t%mask_in)  k = ieor(t%mask, k)
       t%branch(k)%set = .true.
       t%n_branches = t%n_branches+1
    end do
    do i=0, t%n_externals-1
       k = ibset(0,i)
       if (iand(k, t%mask_in) == t%mask_in)  k = ieor(t%mask, k)
       if (t%branch(ieor(t%mask, k))%set) then
          t%branch(ieor(t%mask, k))%set = .false.
          t%branch(k)%set = .true.
       else if (.not.t%branch(k)%set) then
          t%branch(k)%set = .true.
          t%n_branches = t%n_branches+1
       end if
    end do
    if (t%n_branches /= t%n_externals*2-3) then
       call write(6,t)
       stop 'Inconsistent tree array: wrong number of branches set'
    end if
    do k=1, size(t%branch)
       if (t%branch(k)%set .and. decay_level(k)/=1) then
          call branch_set_relatives(k)
       end if
    end do
  contains
    subroutine branch_set_relatives(k)
      integer(kind=TC), intent(in) :: k
      integer(kind=TC) :: m,n
      do m=1, k-1
         if(iand(k,m)==m) then
            n = ieor(k,m)
            if ( t%branch(m)%set .and. t%branch(n)%set ) then
               t%branch(k)%daughter(1) = m;  t%branch(k)%daughter(2) = n
               t%branch(m)%mother      = k;  t%branch(n)%mother      = k
               t%branch(m)%sibling     = n;  t%branch(n)%sibling     = m
               return
            end if
         end if
      end do
      call write(6, t)
      stop 'Inconsistent tree array: Missing daughter branch(es)'
    end subroutine branch_set_relatives

  end subroutine tree_from_array

  subroutine tree_to_array(t,a)
    type(decay_tree), intent(in) :: t
    integer(kind=TC), dimension(:), intent(inout) :: a
    integer(kind=TC) :: i,k
    i=0
    do k=1, size(t%branch)
       if (t%branch(k)%set) then
          i = i+1
          a(i) = k
       end if
    end do
  end subroutine tree_to_array
  subroutine flip_t_to_s_channel(t)
    type(decay_tree), intent(inout) :: t
    integer(kind=TC) :: k, f, m, n, d, s
    if (t%n_in == 2) then
       FLIP: do k=3, t%mask-1
          if (.not. t%branch(k)%set) cycle FLIP
          f = iand(k,t%mask_in)
          if (f==0 .or. f==k) cycle FLIP
          m = t%branch(k)%mother
          s = t%branch(k)%sibling
          if (s==0) call find_orphan(s)
          d = t%branch(k)%daughter(1)
          n = ior(d,s)  
          t%branch(k)%set = .false.
          t%branch(n)%set = .true.
          t%branch(n)%origin = k
          t%branch(n)%daughter(1) = d; t%branch(d)%mother  = n
          t%branch(n)%daughter(2) = s; t%branch(s)%mother  = n
          t%branch(d)%sibling = s;  t%branch(s)%sibling = d
          t%branch(n)%sibling = f;  t%branch(f)%sibling = n
          t%branch(n)%mother      = m
          t%branch(f)%mother      = m
          if (m/=0) then
             t%branch(m)%daughter(1) = n
             t%branch(m)%daughter(2) = f
          end if
          !if (n/=t%mask_out) t%branch(n)%friend = f
          t%branch(n)%friend = f
       end do FLIP
    end if
  contains
    subroutine find_orphan(s)
      integer(kind=TC) :: s
      do s=1, t%mask_out
         if (t%branch(s)%set .and. t%branch(s)%mother==0) return
      end do
      call write(6, t)
      stop "Inconsistent tree array: can't flip t channel"
    end subroutine find_orphan
  end subroutine flip_t_to_s_channel

  function flipped(t, kt) result(ks)
    type(decay_tree), intent(in) :: t
    integer(kind=TC), intent(in) :: kt
    integer(kind=TC) :: ks
    if (iand(kt,t%mask_in)==0) then
       ks = kt
    else
       ks = t%branch(iand(kt,t%mask_out))%mother
    end if
  end function flipped


end module decay_trees
