! WHIZARD 1.00beta Dec 01 2000
! 
! (C) 1999, 2000 by Wolfgang Kilian <kilian@particle.uni-karlsruhe.de>
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

! For the LEP Monte Carlos, a standard common block has been proposed
! in AKV89.  We strongly recommend its use.  (The description is an
! abbreviated transcription of AKV89, Vol. 3, pp. 327-330).
!
!
! NMXHEP is the maximum number of entries:
!
!
! NEVHEP is normally the event number, but may take special
! values as follows:
!
!    0   the program does not keep track of event numbers.
!   -1   a special initialization record.
!   -2   a special final record.
!
!
! NHEP holds the number of entries for this event.
!
!
! The entry ISTHEP(N) gives the status code for the Nth entry,
! with the following semantics:
!    0       a null entry.
!    1       an existing entry, which has not decayed or fragmented.
!    2       a decayed or fragmented entry, which is retained for
!            event history information.
!    3       documentation line.
!    4- 10   reserved for future standards.
!   11-200   at the disposal of each model builder.
!  201-      at the disposal of users.
!
!
! The Particle Data Group has proposed standard particle codes,
! which are to be stored in IDHEP(N).
!
!
! JMOHEP(1,N) points to the Nth entry's mother, if any.
! It is set to zero for initial entries.
! JMOHEP(2,N) points to the second mother, if any.
!
!
! JDAHEP(1,N) and JDAHEP(2,N) point to the Nth entry's first and
! last daughter, if any.  These are zero for entries which have not
! yet decayed.  The other daughters are stored in between these two.
!
!
! In PHEP we store the momentum of the particle, more specifically
! this means that PHEP(1,N), PHEP(2,N), and PHEP(3,N) contain the
! momentum in the x, y, and z direction (as defined by the machine
! people), measured in GeV/c.  PHEP(4,N) contains the energy in GeV
! and PHEP(5,N) the mass in GeV/c**2.  The latter may be negative for
! spacelike partons.
!
!
! Finally VHEP is the place to store the position of the production
! vertex.  VHEP(1,N), VHEP(2,N), and VHEP(3,N) contain the x, y,
! and z coordinate (as defined by the machine people), measured in mm.
! VHEP(4,N) contains the production time in mm/c.
!
!
! As an amendment to the proposed standard common block HEPEVT, we
! also have a polarisation common block HEPSPN, as described in
! AKV89.  SHEP(1,N), SHEP(2,N), and SHEP(3,N) give the x, y, and z
! component of the spinvector s of a fermion in the fermions
! restframe.
!
! By convention, SHEP(4,N) is always 1.

module hepevt_common

  use kinds, only: double !NODEP!
  use lorentz
  use particles
  use events

  implicit none
  private

  public :: hepevt_fill
  public :: hepevt_write

  integer, parameter :: nmxhep = 4000
  integer :: nevhep, nhep
  integer, dimension(nmxhep) :: isthep, idhep
  integer, dimension(2, nmxhep) :: jmohep, jdahep
  real(kind=double), dimension(5, nmxhep) :: phep
  real(kind=double), dimension(4, nmxhep) :: vhep
  common /HEPEVT/ nevhep, nhep, isthep, idhep, &
       & jmohep, jdahep, phep, vhep

  interface hepevt_write
     module procedure hepevt_write_unit
  end interface

contains

  subroutine hepevt_fill(evt)
    type(event), intent(in) :: evt
    integer :: i
    real(kind=double), dimension(0:3) :: p
    nevhep = evt%count
    nhep = evt%n_out
    do i=1, nhep
       isthep(i) = 1
       idhep(i) = particle_code(evt%prt(i))
       p = array(particle_momentum(evt%prt(i)))
       phep(1:3, i) = p(1:3)
       phep(4, i) = p(0)
       phep(5, i) = particle_mass(evt%prt(i))
    end do
  end subroutine hepevt_fill

  subroutine hepevt_write_unit(u)
    integer, intent(in) :: u
    integer :: i
    write(u, *) 'HEPEVT contents:'
    select case(nevhep)
    case(-1)
       write(u, *) 'Initialization record (nevhep = -1)'
    case(-2)
       write(u, *) 'Finalization record (nevhep = -2)'
    case(0:)
       write(u, *) 'Event number:      ', nevhep
       write(u, *) 'Number of entries: ', nhep
       do i=1, nhep
          write(u, *) 'Entry #', i, ':'
          write(u, *) '  Status code: ', isthep(i)
          write(u, *) '  Particle: ', particle_name(idhep(i))
          write(u, *) '  Particle ID: ', idhep(i)
          write(u, *) '  Momentum: ', phep(1:3, i)
          write(u, *) '  Energy:   ', phep(4, i)
          write(u, *) '  Mass:     ', phep(5, i)
       end do
    case default
       write(u, *) 'Illegal record'
    end select
  end subroutine hepevt_write_unit


end module hepevt_common
