! WHIZARD 1.00beta Dec 01 2000
! 
! (C) 1999, 2000 by Wolfgang Kilian <kilian@particle.uni-karlsruhe.de>
!
! WHIZARD is free software; you can redistribute it and/or modify it
! under the terms of the GNU General Public License as published by 
! the Free Software Foundation; either version 2, or (at your option)
! any later version.
!
! WHIZARD is distributed in the hope that it will be useful, but
! WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
! GNU General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this program; if not, write to the Free Software
! Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
!
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This file has been stripped of most comments.  For documentation, refer
! to the source 'whizard.nw'

module vamp_interface

  use kinds, only: default !NODEP!
!  use kinds, only: double !NODEP!
  use limits, only: BUFFER_SIZE, WHIZARD_ROOT, GRID_EXTENSION_LENGTH
  use file_utils, only: free_unit !NODEP!
  use mpi90 !NODEP!
  use lorentz, only: array
  use particles
  use polarization
  use events
  use hepevt_common
  use decay_forests, only: decay_forest, decay_grove
  use cuts
  use phase_space, only: event_fill, map_decay_tree_using_grid
  use beams, only: structure_function
  use monte_carlo_states
  use process_interface !NODEP!
!  use process_interface, only: allow_helicities, sqme  !NODEP!
  use tao_random_numbers !NODEP!
  use vamp  !NODEP!
  !use vampi !NODEP!

  implicit none
  private

  public :: mc_integrate
  public :: mc_get_accuracy
  public :: mc_generate 

contains

  subroutine mc_integrate(mci, filename, time_per_event)
    type(monte_carlo_input), intent(in) :: mci
    character(len=*), intent(in), optional :: filename
    real(kind=default), intent(out), optional :: time_per_event
    real(kind=default), dimension(2,mcs%f%n_dim) :: region
    real(kind=default) :: int_best, err_best, chi_best, eff_best, acc_best
    logical :: improved_grid, skip_integration, ok1, ok2
    integer :: level, it, it0, it1
    integer :: it_file, it_best
    integer :: time0, time1, tick_rate, proc_id
    character :: lext
    integer, parameter :: itext_len = GRID_EXTENSION_LENGTH
    character(len=itext_len) :: itext
    character(len=*), parameter :: &
         & itext_fmt = "(i" // char(ichar('0')+itext_len) // &
         &             "." // char(ichar('0')+itext_len) // ")"
    call mpi90_rank (proc_id)
    region(1,:)  = 0;  region(2,:)  = 1
    if (any(mcs%weights(1,:)>0) .and. all(mcs%weights(1,:)>=0)) &
         & mcs%weights(1,:) = mcs%weights(1,:) / sum(mcs%weights(1,:))
    call vamp_create_grids(mcs%g, region, mcs%calls(2,1), mcs%weights(1,:), &
         & stratified=mcs%stratified)
    call vamp_create_grids(mcs%g_best, region, mcs%calls(2,1), &
         & mcs%weights(1,:), stratified=mcs%stratified)
    mcs%actual_integration_done = .false.
    mcs%checksum = checksum(mcs, 0)
    skip_integration = .false.
    int_best = 0
    err_best = 0
    chi_best = 0
    eff_best = 0
    acc_best = huge(real(1.,kind=default))
    it_file = 0
    if (present(filename) .and. mcs%read_grids ) then
       if (proc_id==WHIZARD_ROOT) then
          if (mcs%read_grids_force) then
             call mc_read_grids &
                  & (mcs%g_best, filename//".grb", it_file, it_best, &
                  &  int_best, err_best, chi_best, eff_best, acc_best, &
                  &  mcs%logtable, ok2)
             call mc_read_grids &
                  & (mcs%g, filename//".grc", it_file, it, &
                  &  mcs%integral, mcs%error, mcs%chi2, &
                  &  mcs%efficiency, mcs%accuracy, &
                  &  mcs%logtable, ok1)
          else
             call mc_read_grids &
                  & (mcs%g_best, filename//".grb", it_file, it_best, &
                  &  int_best, err_best, chi_best, eff_best, acc_best, &
                  &  mcs%logtable, ok2)
             call mc_read_grids &
                  & (mcs%g, filename//".grc", it_file, it, &
                  &  mcs%integral, mcs%error, mcs%chi2, &
                  &  mcs%efficiency, mcs%accuracy, mcs%logtable, ok1)
          end if
          if (ok1 .and. .not.ok2) then
             call vamp_copy_grids (mcs%g_best, mcs%g)
             int_best = mcs%integral
             err_best = mcs%error
             chi_best = mcs%chi2
             eff_best = mcs%efficiency
             acc_best = mcs%accuracy
             it_best  = it
          else if (ok2 .and. .not.ok1) then   
             call vamp_copy_grids (mcs%g, mcs%g_best)
             mcs%integral = int_best
             mcs%error = err_best
             mcs%chi2 = chi_best
             mcs%efficiency = eff_best
             mcs%accuracy = acc_best
             it = it_best
          end if
       end if
       call vamp_broadcast_grids(mcs%g, WHIZARD_ROOT)
    else
       skip_integration = .false.
    end if
    ITERATE: do level=1,4
       if (level>3 .and. mcs%calls(1,level)*mcs%calls(2,level)==0) exit ITERATE
       lext = char(ichar('0')+level)
       it1 = sum(mcs%calls(1,:level))
       if (level==2) then
          it0 = mcs%calls(1,1)+1
       else
          it0 = it1
       end if
       if (level>1) then
          call vamp_copy_grids(mcs%g, mcs%g_best)
          call vamp_discard_integrals &
               & (mcs%g, mcs%calls(2,level), stratified=mcs%stratified)
       end if
       mcs%weights(mcs%it+1,:) = mcs%g%weights
       if (present(filename) .and. proc_id==WHIZARD_ROOT) then
          call read(filename//".cut"//lext, mcs%cc, trim(mcs%process_id), level)
          if (mcs%screen_results) call write(6, mcs%cc, level)
       end if
       call mpi90_broadcast(mcs%cc, WHIZARD_ROOT)
       if (level==1 .and. mcs%screen_results .and. proc_id==WHIZARD_ROOT) then
          call mc_write_header(6, trim(mcs%process_id), mcs%checksum)
       end if
       select case(level)
       case(1,3)
          skip_integration =  sum(mcs%calls(1,:level)) <= it_file
       case default
          skip_integration = .false.
       end select
       if (mcs%screen_results .and. PROC_ID==0) then
          select case(level)
          case(1)
             write(6,'(1x,A)',advance='no') '! Adapting (fixed weights): '
          case(2) 
             write(6,'(1x,A)',advance='no') '! Adapting (var. weights): '
          case(3) 
             write(6,'(1x,A)',advance='no') '! Integrating (fixed w.): '
          case(4) 
             write(6,'(1x,A)',advance='no') '! Preparing grid: '
          end select
          if (mcs%calls(1,level)==1) then
             write(6,*) 'Generating', mcs%calls(1,level), 'sample of', &
                  & mcs%calls(2,level), 'events ...'
          else
             write(6,*) 'Generating', mcs%calls(1,level), 'samples of', &
                  & mcs%calls(2,level), 'events ...'
          end if
       end if
       if (present(time_per_event)) call system_clock(time0)
       select case(level)
       case(1,3)
          if (skip_integration) then
             if (mcs%screen_results .and. proc_id==WHIZARD_ROOT) then
                it = sum(mcs%calls(1,:level))
                call mc_write_result(6, 1, mcs%logtable(it:it))
                call mc_write_hline(6)
             end if
          else
             call vamp_sample_grids &
                  & (mcs%rng, mcs%g, mc_sample, mcs%calls(1,level), &
                  &  integral=mcs%integral, &
                  &  std_dev=mcs%error, avg_chi2=mcs%chi2, &
                  &  history=mcs%history(mcs%it+1:), &
                  &  histories=mcs%histories(mcs%it+1:,:))
             mcs%actual_integration_done = .true.
             mcs%efficiency = mc_get_efficiency (mcs%g)
             mcs%accuracy = mc_get_accuracy &
                  & (mcs%integral, mcs%error, mcs%calls(1,level)*mcs%calls(2,level))
             if (mcs%use_efficiency) then
                improved_grid = (mcs%efficiency > eff_best)
             else
                improved_grid = (mcs%accuracy < acc_best)
             end if
             if (improved_grid) then
                call vamp_copy_grids(mcs%g_best, mcs%g)
                int_best = mcs%integral
                err_best = mcs%error
                chi_best = mcs%chi2
                eff_best = mcs%efficiency
                acc_best = mcs%accuracy
             end if
             call mc_write_result &
                  & (mcs%logtable(mcs%it + mcs%calls(1,level)), &
                  &  mcs%it + mcs%calls(1,level), &
                  &  mcs%calls(2,level)*mcs%calls(1,level), &
                  &  mcs%integral, mcs%error, mcs%efficiency, mcs%accuracy, &
                  &  use_efficiency=mcs%use_efficiency, &
                  &  improved=improved_grid, chi2=mcs%chi2)
             if (mcs%screen_results .and. proc_id==WHIZARD_ROOT) then
                call mc_write_result &
                     & (6, mcs%it + mcs%calls(1,level), &
                     &  mcs%calls(2,level)*mcs%calls(1,level), &
                     &  mcs%integral, mcs%error, mcs%efficiency, mcs%accuracy, &
                     &  use_efficiency=mcs%use_efficiency, &
                     &  improved=improved_grid, chi2=mcs%chi2)
                call mc_write_hline(6)
             end if
          end if
          select case (level)
          case(1)
             do it=mcs%it + 1, mcs%it + mcs%calls(1,level)
                mcs%weights(it,:) = mcs%g%weights
             end do
          case(3)
             do it=mcs%it + 1, mcs%it + mcs%calls(1,level)
                mcs%weights(it,:) = mcs%g%weights
             end do
          end select
          mcs%it = mcs%it + mcs%calls(1,level)
          if (.not.skip_integration) then
             if (improved_grid) it_best = mcs%it
             if (mcs%write_grids) &
                  & call mc_write_grids &
                  &   (mcs%g_best, "whizard.grb", mcs%it, it_best, &
                  &    int_best, err_best, chi_best, eff_best, acc_best, &
                  &    mcs%logtable(:mcs%it), trim(mcs%process_id), mcs%checksum)
             if (mcs%write_grids) then
                call mc_write_grids &
                     &   (mcs%g, "whizard.grc", mcs%it, mcs%it, &
                     &    mcs%integral, mcs%error, mcs%chi2, &
                     &    mcs%efficiency, mcs%accuracy, &
                     &    mcs%logtable(:mcs%it), trim(mcs%process_id), mcs%checksum)
             end if
             if (mcs%write_all_grids) then
                write(itext, itext_fmt) mcs%it
                call mc_write_grids &
                     &   (mcs%g, "whizard.gr"//trim(itext), mcs%it, mcs%it, &
                     &    mcs%integral, mcs%error, mcs%chi2, &
                     &    mcs%efficiency, mcs%accuracy, &
                     &    mcs%logtable(:mcs%it), trim(mcs%process_id), mcs%checksum)
             end if
             if (mcs%write_logfile)  call write ("whizard.out", mcs, mci)
          end if
       case(2)
          do it=mcs%it+1, mcs%it+mcs%calls(1,level)
             skip_integration =  it <= it_file
             if (skip_integration) then
                if (mcs%screen_results .and. proc_id==WHIZARD_ROOT) then
                   if (len_trim(mcs%logtable(it))>0) then
                      call mc_write_result (6, 1, mcs%logtable(it:it))
                   end if
                end if
             else
                call vamp_sample_grids &
                     & (mcs%rng, mcs%g, mc_sample, 1, &
                     &  integral=mcs%integral, &
                     &  std_dev=mcs%error, avg_chi2=mcs%chi2, &
                     &  history=mcs%history(it:), &
                     &  histories=mcs%histories(it:,:))
                mcs%actual_integration_done = .true.
                mcs%efficiency = mc_get_efficiency (mcs%g)
                mcs%accuracy = mc_get_accuracy &
                     & (mcs%integral, mcs%error, mcs%calls(2,level))
                if (mcs%use_efficiency) then
                   improved_grid = (mcs%efficiency > eff_best)
                else
                   improved_grid = (mcs%accuracy < acc_best)
                end if
                if (improved_grid) then
                   call vamp_copy_grids(mcs%g_best, mcs%g)
                   int_best = mcs%integral
                   err_best = mcs%error
                   chi_best = mcs%chi2
                   eff_best = mcs%efficiency
                   acc_best = mcs%accuracy
                end if
                call mc_write_result &
                     & (mcs%logtable(it), it, mcs%calls(2,level), &
                     &  mcs%integral, mcs%error, mcs%efficiency, mcs%accuracy, &
                     &  use_efficiency=mcs%use_efficiency, improved=improved_grid)
                if (mcs%screen_results .and. proc_id==WHIZARD_ROOT) then
                   call mc_write_result &
                        & (6, it, mcs%calls(2,level), &
                        &  mcs%integral, mcs%error, mcs%efficiency, mcs%accuracy, &
                        &  use_efficiency=mcs%use_efficiency, improved=improved_grid)
                end if
             end if
             mcs%weights(it,:) = mcs%g%weights
             if (it==mcs%apply_mappings) then
                print *, "Scanning grids for mappings ..."
                call mc_endpoint_mappings (mcs%f, mcs%g)
                print *, "Scanning finished."
             end if
             mcs%it = mcs%it + 1
             if (.not.skip_integration) then
                if (improved_grid) it_best = mcs%it
                if (mcs%write_grids) &
                     & call mc_write_grids &
                     &   (mcs%g_best, "whizard.grb", mcs%it, it_best, &
                     &    int_best, err_best, chi_best, eff_best, acc_best, &
                     &    mcs%logtable(:mcs%it), trim(mcs%process_id), mcs%checksum)
                if (mcs%write_grids) then
                   call mc_write_grids &
                        &   (mcs%g, "whizard.grc", mcs%it, mcs%it, &
                        &    mcs%integral, mcs%error, mcs%chi2, &
                        &    mcs%efficiency, mcs%accuracy, &
                        &    mcs%logtable(:mcs%it), trim(mcs%process_id), mcs%checksum)
                end if
                if (mcs%write_all_grids) then
                   write(itext, itext_fmt) mcs%it
                   call mc_write_grids &
                        &   (mcs%g, "whizard.gr"//trim(itext), mcs%it, mcs%it, &
                        &    mcs%integral, mcs%error, mcs%chi2, &
                        &    mcs%efficiency, mcs%accuracy, &
                        &    mcs%logtable(:mcs%it), trim(mcs%process_id), mcs%checksum)
                end if
                if (mcs%write_logfile)  call write ("whizard.out", mcs, mci)
                call mc_refine_weights &
                     & (mcs%g, mcs%weights_power, mcs%use_efficiency, &
                     &  mcs%f%n_groves, mcs%f%grove)
             end if
          end do
          if (mcs%screen_results .and. proc_id==WHIZARD_ROOT) then
             if (skip_integration .and. mcs%n_iterations > it_file) then
                call mc_write_result (6, 1, mcs%logtable(it_file:it_file))
             end if
             call mc_write_hline(6)
          end if
       case(4)
          call vamp_warmup_grids &
               & (mcs%rng, mcs%g, mc_sample, mcs%calls(1,level), &
               &  history=mcs%history(mcs%it+1:), &
               &  histories=mcs%histories(mcs%it+1:,:))
          mcs%actual_integration_done = .true.
          mcs%efficiency = mc_get_efficiency(mcs%g)
          if (mcs%screen_results .and. proc_id==WHIZARD_ROOT) then
             call mc_write_result &
                  & (6, it1, mcs%calls(2,level), &
                  &  mcs%integral, mcs%error, mcs%efficiency, mcs%accuracy)
             call mc_write_hline(6)
          end if
          call mc_write_result &
               & (mcs%logtable(it1), it1, mcs%calls(2,level), &
               &  mcs%integral, mcs%error, mcs%efficiency, mcs%accuracy)
          do it=mcs%it + 1, mcs%it + mcs%calls(1,level)
             mcs%weights(it,:) = mcs%weights(1,:)
          end do
          mcs%it = mcs%it + mcs%calls(1,level)
          if (mcs%write_logfile)  call write ("whizard.out", mcs, mci)
       end select
       if (present(time_per_event)) call system_clock(time1, tick_rate)
       if (present(time_per_event) .and. PROC_ID==0) then
          if (tick_rate*mcs%calls(2,level)*mcs%calls(1,level)*mcs%efficiency &
               & /= 0) then
             time_per_event = real((time1-time0)) &
                  &           / real(mcs%calls(2,level)*mcs%calls(1,level)) &
                  &           / real(tick_rate) / mcs%efficiency
          else
             time_per_event = 0
          end if
       end if
       mcs%checksum = checksum(mcs, level)
    end do ITERATE
  end subroutine mc_integrate

  subroutine mc_refine_weights(g, power, use_efficiency, n_groves, grove)
    type(vamp_grids), intent(inout) :: g
    real(kind=default), intent(in) :: power
    logical, intent(in) :: use_efficiency
    integer, intent(in) :: n_groves
    type(decay_grove), dimension(:), intent(in) :: grove
    real(kind=default), dimension(size(g%weights)) :: weights
    real(kind=default) :: w
    integer :: i
    if (use_efficiency) then
       weights = g%weights * mc_get_efficiency_vector (g) ** power
    else
       weights = g%weights * vamp_get_variance (g%grids) ** power
    end if
    do i=1, n_groves
       w = sum(weights(grove(i)%first : grove(i)%last)) &
            &       / (grove(i)%last - grove(i)%first + 1)
       weights(grove(i)%first : grove(i)%last) = w
    end do
    call vamp_update_weights(g, weights)
  end subroutine mc_refine_weights

  subroutine mc_endpoint_mappings (f, g)
    type(decay_forest), intent(inout) :: f
    type(vamp_grids), intent(inout) :: g
    real(kind=default), dimension(:,:), allocatable :: x
    integer, dimension(:), allocatable :: nbins
    integer :: ndim, nbins_max, channel, i
    do channel=1, size(g%grids)
       ndim = size(g%grids(channel)%div)
       allocate (nbins(ndim))
       do i=1, ndim
          nbins(i) = ubound(g%grids(channel)%div(i)%x, 1)
       end do
       nbins_max = maxval(nbins)
       allocate (x(0:nbins_max, ndim))
       do i=1, ndim
          x(0:nbins(i), i) = g%grids(channel)%div(i)%x(0:)
       end do
!       print *, x(0:nbins(2), 2)
!       print *
       call map_decay_tree_using_grid &
            & (f%map(:,channel), f%tree(channel), f%n_dim1, f%n_dim2, &
            &  x, nbins)
       do i=1, ndim
          g%grids(channel)%div(i)%x(0:) = x(0:nbins(i), i)
       end do
       deallocate (x)
       deallocate (nbins)
    end do
  end subroutine mc_endpoint_mappings

  function mc_get_efficiency(g) result(efficiency)
    type(vamp_grids), intent(in) :: g
    real(kind=default) :: efficiency
    real(kind=default) :: wsum
    real(kind=default), dimension(size(g%grids)) :: weights
    where(g%num_calls >= 2 .and. g%grids%f_max>0)
       weights = g%weights * g%grids%f_max
    elsewhere
       weights = 0
    end where
    wsum = sum(weights)
    if (wsum/=0) then
       efficiency = &
            & dot_product(mc_get_efficiency_vector(g), weights) / wsum
    else
       efficiency = 1
    end if
  end function mc_get_efficiency
  function mc_get_efficiency_vector(g) result(efficiency)
    type(vamp_grids), intent(in) :: g
    real(kind=default), dimension(size(g%grids)) :: efficiency
    where(g%num_calls >= 2 .and. g%grids%f_max>0)
       efficiency = g%grids%mu(1) / g%grids%f_max
    elsewhere
       efficiency = 0
    end where
  end function mc_get_efficiency_vector

  function mc_get_accuracy(integral, error, calls) result(a)
    real(kind=default), intent(in) :: integral, error
    integer, intent(in) :: calls
    real(kind=default) :: a
    if (integral/=0) then
       a = error / integral * sqrt(real(calls, kind=default))
    else
       a = 0
    end if
  end function mc_get_accuracy

  subroutine mc_write_grids &
       & (g, name, iterations, it, &
       &  integral, error, chi2, efficiency, accuracy, &
       &  logtable, process_id, checksum)
    type(vamp_grids), intent(in) :: g
    character(len=*), intent(in) :: name, process_id
    integer, intent(in) :: iterations, it
    real(kind=default), intent(in) :: integral, error, chi2
    real(kind=default), intent(in) :: efficiency, accuracy
    character(len=*), dimension(:), intent(in) :: logtable
    integer, intent(in) :: checksum
    integer :: u
    !write(*,*) '! Writing grid data to file ', name
    u = free_unit()
    open(u, file=name, action="write", status="replace")
    write(u, "(1x,a)")   "WHIZARD version 1.00beta (Dec 01 2000)"
    write(u, "(1x,a16,1x,a)")    "process_id     =", process_id
    write(u, "(1x,a16,1x,i15)")  "checksum       =", checksum
    write(u, "(1x,a16,1x,i4)")   "iterations     =", iterations
    write(u, "(1x,a16,1x,i4)")   "this_iteration =", it
    write(u, "(1x,a16,1x,e30.22)") "integral       =", integral
    write(u, "(1x,a16,1x,e30.22)") "error          =", error
    write(u, "(1x,a16,1x,e30.22)") "chi2           =", chi2
    write(u, "(1x,a16,1x,e30.22)") "efficiency     =", efficiency
    write(u, "(1x,a16,1x,e30.22)") "accuracy       =", accuracy
    write(u, "(1x,a)")   "begin logtable"
    call mc_write_result(u, iterations, logtable)
    write(u, "(1x,a)")   "end logtable"
    call vamp_write_grids(g, u)
    close(u)
  end subroutine mc_write_grids
  subroutine mc_read_grids &
       & (g, name, iterations, it, &
       &  integral, error, chi2, efficiency, accuracy, &
       &  logtable, ok, checksum)
    type(vamp_grids), intent(inout) :: g
    character(len=*), intent(in) :: name
    character(len=*), dimension(:), intent(inout) :: logtable
    integer, intent(inout) :: iterations, it
    real(kind=default), intent(inout) :: integral, error, chi2
    real(kind=default), intent(inout) :: efficiency, accuracy
    logical, intent(out) :: ok
    integer, intent(in), optional :: checksum
    character(len=BUFFER_SIZE) :: version
    character :: cdummy
    integer :: u, chsum
    inquire(file=name, exist=ok)
    if (ok) then
       write(*,*) '! Reading grid data from file ', name
       u = free_unit()
       open(u, file=name, action="read", status="old")
       read(u, "(1x,a)") version
       if (present(checksum) &
            & .and. trim(version)/="WHIZARD version 1.00beta (Dec 01 2000)") &
            & stop "WHIZARD version mismatch.  Sorry, this file can't be used."
       read(u, *) cdummy
       read(u, "(1x,a16,1x,i15)") cdummy, chsum
       read(u, "(1x,a16,1x,i4)")  cdummy, iterations
       read(u, "(1x,a16,1x,i4)")  cdummy, it
       read(u, "(1x,a16,1x,e30.22)") cdummy, integral
       read(u, "(1x,a16,1x,e30.22)") cdummy, error
       read(u, "(1x,a16,1x,e30.22)") cdummy, chi2
       read(u, "(1x,a16,1x,e30.22)") cdummy, efficiency
       read(u, "(1x,a16,1x,e30.22)") cdummy, accuracy
       if (present(checksum)) then
          if (chsum /= checksum) call checksum_error
       end if
       read(u, *)
       call mc_read_result(u, iterations, logtable)
       read(u, *)
       call vamp_read_grids(g, u)
       close(u)
    end if
  end subroutine mc_read_grids

  subroutine mc_generate(weight, probability, channel, excess)
    real(kind=default), intent(out), optional :: weight, probability
    integer, intent(in), optional :: channel
    real(kind=default), intent(out), optional :: excess
    real(kind=default), dimension(mcs%f%n_dim) :: x
    real(kind=default) :: wgt
    if (present(channel).and.present(probability)) then
       call vamp_next_event(x, mcs%rng, mcs%g%grids(channel), mc_sample, &
            &               wgt, channel, mcs%g%weights, mcs%g%grids)
       probability = wgt / mcs%g%grids(channel)%f_max
       if (present(weight)) weight = wgt
    else
       call vamp_next_event(x, mcs%rng, mcs%g, mc_sample, phi, excess=excess)
    end if
    call select_flavor_and_helicity &
         & (mcs%evt, mcs%n_hel, &
         &  mcs%n_flv_out, mcs%n_hel_out, mcs%code_out, mcs%hel_state_out, &
         &  mcs%rng)
    mcs%evt%count = mcs%evt%count + 1
    call hepevt_fill (mcs%evt)
  end subroutine mc_generate

  function phi (xi, channel) result (x)
    use kinds !NODEP!
    real(kind=default), dimension(:), intent(in) :: xi
    integer, intent(in) :: channel
    real(kind=default), dimension(size(xi)) :: x
    x = xi
  end function phi

  subroutine select_flavor_and_helicity &
       & (evt, n_hel, n_flv_out, n_hel_out, flv_out, hel_out, rng)
    type(event), intent(inout) :: evt
    integer, intent(in) :: n_hel, n_flv_out, n_hel_out
    integer, dimension(:,:), intent(in) :: flv_out, hel_out
    type(tao_random_state), intent(inout) :: rng
    real(kind=default), dimension(n_flv_out) :: rho_flv
    real(kind=default), dimension(n_hel_out) :: rho_hel
    integer :: f, h, i
    real(kind=default) :: r
    rho_flv = sum(evt%rho_out, dim=1)
    call tao_random_number (rng, r)
    call select (f, rho_flv, n_flv_out, r)
    do i=1, evt%n_out
       call set (evt%prt(i), flv_out(i,f))
    end do
    if (n_hel > 0) then
       rho_hel = evt%rho_out(:,f)
       rho_hel = rho_hel / sum(rho_hel)
       call tao_random_number (rng, r)
       call select (h, rho_hel, n_hel_out, r)
       evt%hel_out = hel_out(:,h)
    else
       evt%hel_out = 0
    end if
  contains
    subroutine select (k, rho, n, r)
      integer, intent(out) :: k
      real(kind=default), dimension(:), intent(in) :: rho
      integer, intent(in) :: n
      real(kind=default), intent(in) :: r
      integer :: i
      real(kind=default) :: rho_sum
      rho_sum = 0
      k = n
      LOOP: do i=1, n
         rho_sum = rho_sum + rho(i)
         if (r<rho_sum) then
            k = i
            exit LOOP
         end if
      end do LOOP
    end subroutine select
  end subroutine select_flavor_and_helicity

  subroutine scatter_set_probabilities (evt, process_id, rho_in, total_factor)
    type(event), intent(inout) :: evt
    character(len=*), intent(in) :: process_id
    real(kind=default), dimension(:,:), intent(in) :: rho_in
    real(kind=default), intent(in) :: total_factor
    real(kind=default), dimension(0:3, evt%n_tot) :: p
    real(kind=default) :: factor_tmp
    integer :: i
    if (evt%scattering) then
       do i=1,2
          p(:,i) = array(particle_momentum(evt%prt(-i)))
       end do
    else
       p(:,1) = array(particle_momentum(evt%prt(0)))
    end if
    do i=1, evt%n_out
       p(:,evt%n_in+i) = array(particle_momentum(evt%prt(i)))
    end do
    call process_scatter (p, rho_in, evt%rho_out, process_id)
    factor_tmp = sum(evt%rho_out)
    if (factor_tmp /= 0) then
       evt%rho_out = evt%rho_out / factor_tmp
    end if
    evt%mc_function_value = factor_tmp * total_factor * evt%flux_factor
  end subroutine scatter_set_probabilities

  function mc_sample(xi, weights, channel, grids) result (s)
    use vamp_grid_type !NODEP!
    real(kind=default), dimension(:), intent(in) :: xi
    real(kind=default), dimension(:), intent(in), optional :: weights
    integer, intent(in), optional :: channel
    type(vamp_grid), dimension(:), intent(in), optional :: grids
    real(kind=default), dimension(mcs%f%n_dim, mcs%f%n_channels) :: rr
    real(kind=default), dimension(2) :: x_parton
    real(kind=default), dimension(mcs%f%n_channels) :: phase_space_factor
    real(kind=default) :: phase_space_volume, strfun_factor
    real(kind=default), dimension(mcs%f%n_channels) :: p_evt, g_xi
    logical, dimension(mcs%f%n_channels) :: active
    real(kind=default) :: dp, s
    integer :: ch
    logical :: ok
    active = (mcs%g%num_calls >= 2)
    rr(:, channel) = xi
    s = 0
    if (all(mcs%beam%fixed_energy)) then
       strfun_factor = 1
       call set_in_state &
            & (mcs%evt, mcs%code(:mcs%n_in,1), mcs%par, mcs%sqrts)
       ok = .true.
    else
       do ch=1, mcs%n_channels
          rr(mcs%f%n_dim1+1:, ch) = rr(mcs%f%n_dim1+1:, channel)
       end do
       call structure_function &
            & (strfun_factor, x_parton, &
            &  rr(mcs%f%n_dim1+mcs%f%n_dim2+1:, channel), &
            &  mcs%beam, ok)
       if (ok) call set_in_state&
            & (mcs%evt, mcs%code(:mcs%n_in,1), mcs%par, mcs%sqrts, &
            &  xi=x_parton, ok=ok)
    end if
    if (ok) then
       call set_out_state(mcs%evt, mcs%code(mcs%n_in+1:,1), mcs%par)
       call event_fill(mcs%evt, rr(:mcs%f%n_dim,:), mcs%f, &
            phase_space_factor, phase_space_volume, channel, active, ok)
       mcs%evt%count_internal = mcs%evt%count_internal + 1
    end if
    if (ok) call apply(mcs%cc, mcs%evt, ok)
    if (ok) then
       where (.not.active)
          phase_space_factor = 0
       end where
       do ch=1, mcs%f%n_channels
          if (active(ch) .and. phase_space_factor(ch)>0) then
             p_evt(ch) = vamp_probability(grids(ch), rr(:,ch))
             g_xi(ch)  = p_evt(ch) / phase_space_factor(ch)
          else
             p_evt(ch) = 0
             g_xi(ch) = 0
          end if
       end do
       dp = dot_product(weights, g_xi)
       if (dp>0) then
          call scatter_set_probabilities &
               & (mcs%evt, trim(mcs%process_id), mcs%rho_in, &
               &  phase_space_volume &
               & * strfun_factor &
               & * p_evt(channel) / dp)
          s = mcs%evt%mc_function_value
          if (any(mcs%evt%rho_out<0)) then
          !    call write(6, mcs%evt)
          !    print *
          !    print *, 'phase space volume = ', phase_space_volume
          !    print *, 'strfun_factor      = ', strfun_factor
          !    print *, 'channel weight     = ', p_evt(channel) / dp
          !    print *
          !    stop 'Negative value of sampling function'
             print *, "! *** Warning: Event #", mcs%evt%count, &
                  & ' has negative entry in probability matrix:'
             print *, mcs%evt%rho_out
             mcs%evt%rho_out = 0
          end if
       end if
    end if
! test !
!    call write(6,mcs%evt); stop
! test !
  end function mc_sample


end module vamp_interface
